import datetime
import logging

from django.utils import timezone
from django.contrib.auth import get_user_model

from django_auth_policy import settings as dap_settings
from django_auth_policy import signals
from django_auth_policy.models import LoginAttempt, PasswordChange


logger = logging.getLogger(__name__)


def disable_expired_users():
    """ Disable users that have been expired

    Run inside the authentication form or backend, and before calling
    django.contrib.auth.authenticate

    Reactivate user by setting is_active to True and last_login to
    now.
    """
    if dap_settings.INACTIVE_USERS_EXPIRY is None:
        return None

    expire_at = timezone.now() - datetime.timedelta(
        days=dap_settings.INACTIVE_USERS_EXPIRY)

    expired = get_user_model().objects.filter(is_active=True,
                                              last_login__lt=expire_at)

    for user in expired:
        logger.warning('User %s disabled because last login was at %s',
                       user.username, user.last_login)
        # Send signal to be used to alert admins
        signals.user_expired.send(sender=user, user=user)

    expired.update(is_active=False)


def locked_username(username):
    """ Returns whether username is locked-out
    """
    if dap_settings.FAILED_AUTH_USERNAME_MAX is None:
        return False

    try:
        last_login = LoginAttempt.objects.filter(username=username
                                                 ).order_by('-id')[0]
    except IndexError:
        # No login attempts for this username and thus no lockout
        return False

    user_lockout = LoginAttempt.objects.filter(username=username,
                                               successful=False,
                                               lockout=True)
    if dap_settings.FAILED_AUTH_PERIOD is not None:
        lockout_count_from = timezone.now() - datetime.timedelta(
            seconds=dap_settings.FAILED_AUTH_PERIOD)
        user_lockout = user_lockout.filter(timestamp__gt=lockout_count_from)

    return (user_lockout.count() >= dap_settings.FAILED_AUTH_USERNAME_MAX and
            last_login.within_lockout_period)


def locked_remote_addr(remote_addr):
    """ Returns whether remote address is locked-out
    """
    if dap_settings.FAILED_AUTH_ADDRESS_MAX is None:
        return False

    # Lock per IP address
    try:
        last_login = LoginAttempt.objects.filter(
            source_address=remote_addr).order_by('-id')[0]
    except IndexError:
        # No previous login attempts for this IP address and thus no lockout
        return False

    ip_lockout = LoginAttempt.objects.filter(source_address=remote_addr,
                                             lockout=True)
    if dap_settings.FAILED_AUTH_PERIOD is not None:
        lockout_count_from = timezone.now() - datetime.timedelta(
            seconds=dap_settings.FAILED_AUTH_PERIOD)
        ip_lockout = ip_lockout.filter(timestamp__gt=lockout_count_from)

    return (ip_lockout.count() >= dap_settings.FAILED_AUTH_ADDRESS_MAX and
            last_login.within_lockout_period)


def enforce_password_change(user):
    """ Checks if the user must change its password

    Returns a tuple with three booleans:
    * password change is required;
    * current password has expired;
    * current password is temporary;
    """
    enforce_change = False
    is_expired = False
    is_temporary = False
    # Retrieve last successful password change for user
    try:
        last_pw_change = PasswordChange.objects.filter(
            user=user, successful=True).order_by('-id')[0]
    except IndexError:
        # No password changes recorded
        # Enforce a password change when user has no PasswordChange
        enforce_change = not dap_settings.ALLOW_EMPTY_PASSWORD_HISTORY

        return (enforce_change, is_expired, is_temporary)

    # Enforce temporary password change
    if last_pw_change.is_temporary:
        enforce_change = True
        is_temporary = True

    # Enforce password expiry policy
    if dap_settings.MAX_PASSWORD_AGE is not None:

        expire_at = last_pw_change.timestamp + datetime.timedelta(
            days=dap_settings.MAX_PASSWORD_AGE)

        if timezone.now() > expire_at:
            enforce_change = True
            is_expired = True

    return (enforce_change, is_expired, is_temporary)



import logging
from collections import OrderedDict

from django import forms
from django.utils.translation import ugettext as _
from django.contrib.auth import authenticate
from django.contrib.auth.forms import (AuthenticationForm, SetPasswordForm,
                                       PasswordChangeForm)

from django_auth_policy.models import PasswordChange, LoginAttempt
from django_auth_policy.validators import (password_min_length,
                                           password_complexity)
from django_auth_policy.checks import (disable_expired_users, locked_username,
                                       locked_remote_addr)


logger = logging.getLogger(__name__)


class StrictAuthenticationForm(AuthenticationForm):

    error_messages = dict(AuthenticationForm.error_messages, **{
        'username_locked_out': _('Your account has been locked. Contact your '
                                 'user administrator for more information.'),
        'address_locked_out': _('Your account has been locked. Contact your '
                                'user administrator for more information.')})

    def __init__(self, request, *args, **kwargs):
        """ Make request argument required
        """
        return super(StrictAuthenticationForm, self).__init__(request, *args,
                                                              **kwargs)

    def clean(self):
        username = self.cleaned_data.get('username')
        password = self.cleaned_data.get('password')
        remote_addr = self.request.META['REMOTE_ADDR']

        logger.info('Authentication attempt, username=%s, address=%s',
                    username, remote_addr)

        if not username and not password:
            return self.cleaned_data

        attempt = LoginAttempt(
            username=username,
            source_address=remote_addr,
            hostname=self.request.get_host()[:100],
            successful=False,
            lockout=True)

        if not username:
            logger.warning(u'Authentication failure, address=%s, '
                           'no username supplied.',
                           remote_addr)
            attempt.save()
            return self.cleaned_data

        if not password:
            logger.warning(u'Authentication failure, username=%s, '
                           'address=%s, no password supplied.',
                           username, remote_addr)
            attempt.save()
            return self.cleaned_data

        if locked_username(username):
            logger.warning(u'Authentication failure, username=%s, address=%s, '
                           'username locked', username, remote_addr)
            attempt.save()
            raise forms.ValidationError(
                self.error_messages['username_locked_out'],
                'username_locked_out')

        if locked_remote_addr(remote_addr):
            logger.warning(u'Authentication failure, username=%s, address=%s, '
                           'address locked', username, remote_addr)
            attempt.save()
            raise forms.ValidationError(
                self.error_messages['address_locked_out'],
                'address_locked_out')

        disable_expired_users()
        self.user_cache = authenticate(username=username,
                                       password=password)
        if self.user_cache is None:
            logger.warning(u'Authentication failure, username=%s, '
                           'address=%s, invalid authentication.',
                           username, remote_addr)
            attempt.save()
            raise forms.ValidationError(
                self.error_messages['invalid_login'] % {
                    'username': self.username_field.verbose_name},
                code='invalid_login')

        if not self.user_cache.is_active:
            logger.warning(u'Authentication failure, username=%s, '
                           'address=%s, user inactive.',
                           username, remote_addr)
            attempt.save()
            raise forms.ValidationError(
                self.error_messages['inactive'],
                code='inactive')

        # Authentication was successful
        logger.info(u'Authentication success, username=%s, address=%s',
                    username, remote_addr)
        attempt.successful = True
        attempt.lockout = False
        attempt.user = self.user_cache
        attempt.save()

        # Reset lockout counts for IP address and username
        LoginAttempt.objects.filter(username=username,
                                    lockout=True).update(lockout=False)
        LoginAttempt.objects.filter(source_address=remote_addr,
                                    lockout=True).update(lockout=False)

        return self.cleaned_data


class StrictSetPasswordForm(SetPasswordForm):
    def clean_new_password1(self):
        pw = self.cleaned_data.get('new_password1')
        if pw:
            password_min_length(pw)
            password_complexity(pw)
        return pw

    def is_valid(self):
        valid = super(StrictSetPasswordForm, self).is_valid()
        if self.is_bound:
            PasswordChange.objects.create(user=self.user, successful=valid,
                                          is_temporary=False)
            if valid:
                logger.info('Password change successful for user %s',
                            self.user)
            else:
                logger.info('Password change failed for user %s',
                            self.user)
        return valid


class StrictPasswordChangeForm(StrictSetPasswordForm, PasswordChangeForm):
    error_messages = dict(StrictSetPasswordForm.error_messages, **{
        'password_unchanged': _("The new password must not be the same as "
                                "the old password"),
        })

    def clean_new_password1(self):
        pw = super(StrictPasswordChangeForm, self).clean_new_password1()

        # Check that old and new password differ
        if (self.cleaned_data.get('old_password') and
                self.cleaned_data['old_password'] == pw):

            raise forms.ValidationError(
                self.error_messages['password_unchanged'],
                'password_unchanged')

        return pw

StrictPasswordChangeForm.base_fields = OrderedDict(
    (k, StrictPasswordChangeForm.base_fields[k])
    for k in ['old_password', 'new_password1', 'new_password2']
)



"""
Django Authentication Policy - Default settings

Warning: lockout of usernames or IP addresses based on number of failed login
attempts can be used by attackers to temporarily lock out valid users.

Alter the settings by adding them to your projects settings file, do not edit
the defaults below.
"""
from django.conf import settings
from django.utils.translation import ugettext as _

# Number of days after which the user is required to change their password
# Set to None to disable this feature
MAX_PASSWORD_AGE = getattr(settings, 'MAX_PASSWORD_AGE', 30)

# By default a password change is enforced when a user has no password change
# history
ALLOW_EMPTY_PASSWORD_HISTORY = getattr(settings,
                                       'ALLOW_EMPTY_PASSWORD_HISTORY', False)

# Number of days without a successful log-in after which a user will be
# disabled. Set to None to disable this feature
INACTIVE_USERS_EXPIRY = getattr(settings, 'INACTIVE_USERS_EXPIRY', 90)

# Number of failed authentications for a specific username within the
# FAILED_AUTH_PERIOD that triggers a lockout (for a 'three strikes and you're
# out' policy)
# Set to None to disable this feature
FAILED_AUTH_USERNAME_MAX = getattr(settings, 'FAILED_AUTH_USERNAME_MAX', 3)

# Number of failed authentications for a specific remote address within the
# FAILED_AUTH_PERIOD that triggers a lockout (for a 'three strikes and you're
# out' policy)
# Set to None to disable this feature
FAILED_AUTH_ADDRESS_MAX = getattr(settings, 'FAILED_AUTH_ADDRESS_MAX', 3)

# Period used to measure the number of failed logins, in seconds
# Set to None to count all failed logins that count towards a lockout
# (Set to None by default)
FAILED_AUTH_PERIOD = getattr(settings, 'FAILED_AUTH_PERIOD', None)

# Failed authentication lockout period in seconds, the default is 3 hours
FAILED_AUTH_LOCKOUT_PERIOD = getattr(settings, 'FAILED_AUTH_LOCKOUT_PERIOD',
                                     60 * 60 * 3)

# Minimum password length requirement
PASSWORD_MIN_LENGTH = getattr(settings, 'PASSWORD_MIN_LENGTH', 10)

# Minimum password length text
PASSWORD_MIN_LENGTH_TEXT = getattr(settings, 'PASSWORD_MIN_LENGTH_TEXT',
    _('Passwords must be at least {length} characters in length.'))

# Additional password strength requirements
# Define a list of character requirements, each requirement defines:
# - 'name': Name of the character class
# - 'chars': string with all characters of that class
# - 'min': minimum number of characters required from this class
# - 'text': a text explaining the complexity rule
#   This text might be prepended with something like: 'Passwords must have',
#   and might be appended with the available characters (between parentheses)
#   Example: Passwords must have at least ... (abcdefghijklmnopqrstuvwxyz)
# Set the whole setting to None to disable additional strength requirements
PASSWORD_COMPLEXITY = getattr(settings, 'PASSWORD_COMPLEXITY', (
    {'name': _('lowercase characters'),
     'chars': 'abcdefghijklmnopqrstuvwxyz',
     'min': 1,
     'text': _('at least one lowercase character')},

    {'name': _('uppercase characters'),
     'chars': 'ABCDEFGHIJKLMNOPQRSTUVWXYZ',
     'min': 1,
     'text': _('at least one uppercase character')},

    {'name': _('numbers'),
     'chars': '0123456789',
     'min': 1,
     'text': _('at least one number')},

    {'name': _('symbols'),
     'chars': '!@#$%^&*()_+-={}[]:;"\'|\\,.<>?/~`',
     'min': 1,
     'text': _('at least one symbol')},
    ))

# Text displayed when password doesn't meet complexity requirements
PASSWORD_COMPLEXITY_TEXT = _('Passwords must have {rule_text}')

# Characters used to generate temporary passwords
TEMP_PASSWORD_CHARS = getattr(settings, 'TEMP_PASSWORD_CHARS',
                              'abcdefghijlkmnopqrstuvwxyz'
                              'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                              '0123456789')

# Temporary password length
TEMP_PASSWORD_LENGTH = getattr(settings, 'TEMP_PASSWORD_LENGTH', 12)

# django_auth_policy replaces the default Django auth UserAdmin to enforce
# authentication policies on the admin interface. Set this to False when
# django_auth_policy shouldn't replace UserAdmin.
REPLACE_AUTH_USER_ADMIN = getattr(settings, 'REPLACE_AUTH_USER_ADMIN', True)
